# Generated by Django 5.2.6 on 2025-09-15 15:49

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('okta_client', '0002_alter_oktauser_options_alter_oktauser_managers_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='oktauser',
            name='last_refresh_timestamp',
            field=models.DateTimeField(blank=True, help_text='The last time the user was updated (refreshed) from Okta', null=True, verbose_name='Last refresh timestamp'),
        ),
        migrations.AddField(
            model_name='oktauser',
            name='okta_activated',
            field=models.DateTimeField(blank=True, help_text='Timestamp of the account activation.', null=True, verbose_name='activated'),
        ),
        migrations.AddField(
            model_name='oktauser',
            name='okta_created',
            field=models.DateTimeField(blank=True, help_text='Timestamp of the account creation.', null=True, verbose_name='created'),
        ),
        migrations.AddField(
            model_name='oktauser',
            name='okta_id',
            field=models.CharField(blank=True, help_text='Okta internal ID for the user.', max_length=50, null=True, unique=True, verbose_name='id'),
        ),
        migrations.AddField(
            model_name='oktauser',
            name='okta_status',
            field=models.CharField(blank=True, choices=[('STAGED', 'Staged'), ('PROVISIONED', 'Pending User Action'), ('ACTIVE', 'Active'), ('RECOVERY', 'Password Reset'), ('PASSWORD_EXPIRED', 'Password Expired'), ('LOCKED_OUT', 'Locked Out'), ('SUSPENDED', 'Suspended'), ('DEPROVISIONED', 'Deprovisioned')], help_text='Status of the Okta account.', max_length=30, verbose_name='status'),
        ),
        migrations.AddField(
            model_name='oktauser',
            name='okta_status_changed',
            field=models.DateTimeField(blank=True, help_text='Timestamp of the last update of the "status" attribute.', null=True, verbose_name='status changed'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='city',
            field=models.CharField(blank=True, help_text="City or locality component of user's address (locality)", max_length=100, verbose_name='city'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='costCenter',
            field=models.CharField(blank=True, help_text='Name of a cost center assigned to user', max_length=100, verbose_name='cost center'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='countryCode',
            field=models.CharField(blank=True, help_text="Country name component of user's address (country)", max_length=2, verbose_name='country code'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='date_joined',
            field=models.DateTimeField(auto_now_add=True, help_text='The timestamp when the local account was created', verbose_name='date joined'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='department',
            field=models.CharField(blank=True, help_text="Name of user's department", max_length=100, verbose_name='department'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='displayName',
            field=models.CharField(blank=True, help_text='Name of the user, suitable for display to end users', max_length=250, verbose_name='display name'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='division',
            field=models.CharField(blank=True, help_text="Name of user's division", max_length=100, verbose_name='division'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='email',
            field=models.EmailField(help_text='Primary email address of user', max_length=254, validators=[django.core.validators.RegexValidator(message='5 <= value length <= 100', regex='^.{5,100}$')], verbose_name='email'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='employeeNumber',
            field=models.CharField(blank=True, help_text='Organization or company assigned unique identifier for the user', max_length=100, verbose_name='employee number'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='firstName',
            field=models.CharField(help_text='Given name of the user (givenName)', max_length=50, verbose_name='first name'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='honorificPrefix',
            field=models.CharField(blank=True, help_text='Honorific prefix(es) of the user, or title in most Western languages', max_length=50, verbose_name='honorific prefix'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='honorificSuffix',
            field=models.CharField(blank=True, help_text='Honorific suffix(es) of the user', max_length=50, verbose_name='honorific suffix'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='is_staff',
            field=models.BooleanField(default=False, help_text='Designates whether the user can log into the admin site.', verbose_name='staff status'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='lastName',
            field=models.CharField(help_text='Family name of the user (familyName)', max_length=50, verbose_name='last name'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='locale',
            field=models.CharField(blank=True, help_text="User's default location for purposes of localizing items such as currency, date time format, numerical representations, etc.", max_length=5, verbose_name='locale'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='login',
            field=models.CharField(help_text='Unique identifier for the user (username)', max_length=100, primary_key=True, serialize=False, validators=[django.core.validators.RegexValidator(message='5 <= value length <= 100', regex='^.{5,100}$')], verbose_name='login'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='manager',
            field=models.CharField(blank=True, help_text="DisplayName of the user's manager", max_length=100, verbose_name='manager'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='managerId',
            field=models.CharField(blank=True, help_text="ID of a user's manager", max_length=100, verbose_name='manager ID'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='middleName',
            field=models.CharField(blank=True, help_text='Middle name(s) of the user', max_length=50, verbose_name='middle name'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='mobilePhone',
            field=models.CharField(blank=True, help_text='Mobile phone number of user', max_length=100, verbose_name='mobile phone'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='nickName',
            field=models.CharField(blank=True, help_text='Casual way to address the user in real life', max_length=50, verbose_name='nickname'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='organization',
            field=models.CharField(blank=True, help_text="Name of user's organization", max_length=100, verbose_name='organization'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='postalAddress',
            field=models.CharField(blank=True, help_text="Mailing address component of user's address", max_length=100, verbose_name='postal address'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='preferredLanguage',
            field=models.CharField(blank=True, help_text="User's preferred written or spoken languages", max_length=100, verbose_name='preferred language'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='primaryPhone',
            field=models.CharField(blank=True, help_text='Primary phone number of user such as home number', max_length=100, verbose_name='primary phone'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='profileUrl',
            field=models.URLField(blank=True, help_text="URL of user's online profile (e.g. a web page)", verbose_name='profile URL'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='secondEmail',
            field=models.EmailField(blank=True, help_text='Secondary email address of user typically used for account recovery', max_length=254, validators=[django.core.validators.RegexValidator(message='5 <= value length <= 100 (or nothing)', regex='(^.{5,100}$)|(ˆ$)')], verbose_name='second email'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='state',
            field=models.CharField(blank=True, help_text="State or region component of user's address (region)", max_length=100, verbose_name='state'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='streetAddress',
            field=models.CharField(blank=True, help_text="Full street address component of user's address", max_length=100, verbose_name='street address'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='timezone',
            field=models.CharField(blank=True, help_text="User's time zone", max_length=100, verbose_name='time zone'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='title',
            field=models.CharField(blank=True, help_text='User\'s title, such as "Vice President"', max_length=100, verbose_name='title'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='userType',
            field=models.CharField(blank=True, help_text='Used to describe the organization to user relationship such as "Employee" or "Contractor"', max_length=100, verbose_name='user type'),
        ),
        migrations.AlterField(
            model_name='oktauser',
            name='zipCode',
            field=models.CharField(blank=True, help_text="Zipcode or postal code component of user's address (postalCode)", max_length=100, verbose_name='ZIP code'),
        ),
    ]



"""
Utility functions
"""

from logging import getLogger

LOGGER = getLogger(__name__)


def report_signal_results(results, signal_text):
	"""Report signal results
	To process the result of sending signals that expect no return.
	"""

	for handler, result in results:
		handler = '.'.join((handler.__module__, handler.__name__))
		if isinstance(result, Exception):
			LOGGER.error(f'Handler "%s" experienced an error while processing the "{signal_text}" signal: %s', handler, result)
		elif result:
			LOGGER.warning(f'Handler "%s" returned a value while processing the "{signal_text}" signal. The calling code does not expect an answer, discarding: %s', handler, result)
		else:
			LOGGER.debug(f'Handler "%s" completed successfully the processing of the "{signal_text}" signal.', handler)



#python
"""
Okta authentication backend for Django.
"""

from logging import getLogger

from django.contrib.auth import get_user_model
from django.contrib.auth.backends import RemoteUserBackend

from asgiref.sync import async_to_sync
from rest_framework.authentication import TokenAuthentication as DjangoRESTTokenAuthentication

from .api_client import OktaAPIClient
from .groups import set_user_groups

LOGGER = getLogger(__name__)
UserModel = get_user_model()


class OktaSAMLBackend(RemoteUserBackend):
	"""Okta auth backend
	Include Okta related specifics to the authentication process.
	"""
	
	create_unknown_user = True

	def __getattr__(self, name):
		"""Lazy instantiation
		It provides a mechanism for lazy instantiation of the Okta API client and some other things.

		:param name: The name of the attribute being accessed.
		:type name: str
		:returns: the attribute value
		"""

		if name == '_api_client':
			value = OktaAPIClient()
		else:
			return getattr(super(), name)
		self.__setattr__(name, value)
		return value

	def authenticate(self, request, login, **saml_attributes):
		"""Authenticate user
		Use the provided login to create a local account. It will try to update the attributes and group membership from Okta if the API client is configured. It will apply the SAML attributes if present. Since this is part of the SAML authentication, it will never return None, it will always create the local user matching the provided "login", since "create_unknown_user" is set to True.

		:param request: the request object, unused so far
		:type request: DjangoHTTPRequest
		:param login: the user login to "authenticate"
		:type login: str
		:param saml_attributes: the SAML attributes, if any
		:type saml_attributes: dict
		:return: the authenticated user
		:rtype: UserModel
		"""

		user = super().authenticate(request, login)
		if saml_attributes:
			groups = saml_attributes.pop('groups', [])
			LOGGER.debug('Updating user with SAML attributes: %s <- %s', login, saml_attributes)
			user.update(**saml_attributes)
			user.save()
			if groups:
				set_user_groups(user, groups)
		return user

	def configure_user(self, request, user, created=True):
		"""Configure the user
		User attributes and group membership are updated from Okta. It will only do so if the Users API endpoint is available and the user is outdated.

		:param request: the request object, not used so far
		:type request: DjangoHTTPRequest
		:param user: the user to configure
		:type user: UserModel
		:param created: if the "authenticate" method created the local record
		:type created: bool
		:return: the configured user
		:rtype: UserModel
		"""

		if user.is_outdated and async_to_sync(self._api_client.ping_users_endpoint)():
			async_to_sync(user.update_from_okta)()
			async_to_sync(user.set_groups_from_okta)()
		return user

	def user_can_authenticate(self, remote_user):
		"""Returns whether the user is allowed to authenticate
		This backend should be used with SAML, which is a federated authentication scheme, so there shouldn't be a local permission check for authentication.

		:param remote_user: the user to check for
		:type remote_user: str
		:return: always True
		:rtype: bool
		"""

		return True


class DjangoRESTBearerTokenAuthentication(DjangoRESTTokenAuthentication):
	"""Token authentication for REST
	Like the builtin rest_framework.authentication.TokenAuthentication method but using the "Bearer" word instead of the default "Token" (it simplifies the compatibility with Postman)
	"""

	keyword = 'Bearer'


"""
Logic for membership of Django Groups.
"""

from logging import getLogger

from django.contrib.auth.models import Group

from .signals.groups import group_created, user_joined_group, user_left_group
from .utils import report_signal_results

LOGGER = getLogger(__name__)


def join_group(user, group):
	"""Join a group
	Add a user to a group. Since the group system is not pluggable, any customization to it should be done elsewhere.

	:param user: the user instance
	:type user: object
	:param group: the group instance or something that can be used to retrieve it
	:type group: object|str
	"""

	if not isinstance(group, Group):
		group, created = Group.objects.get_or_create(name=group)
		if created:
			LOGGER.debug('Created new local group: %s', group)
			group_created.send(sender=group)
	LOGGER.debug('Group %s is getting a new member: %s', group, user)
	group.user_set.add(user)


def leave_group(user, group):
	"""Leave a group
	Removes the user from a group. Since the group system is not pluggable, any customization to it should be done elsewhere.

	:param user: the user instance
	:type user: object
	:param group: the group instance or something that can be used to retrieve it
	:type group: object|str
	"""

	if not isinstance(group, Group):
		try:
			group = Group.get(name=group)
		except Group.DoesNotExist:
			LOGGER.debug("Unexisting group %s couldn't lose user: %s", group, user)
			return

	LOGGER.debug('Group %s is losing a member: %s', group, user)
	group.user_set.remove(user)


def set_group_members(group, users):
	"""Replace the group members
	Since the group system is not pluggable, any customization to it should be done elsewhere.

	:param group: the group instance or the name (to create it)
	:type group: Group|str
	:param users: a list of user instances
	:type users: list[UserModel]
	"""

	if not isinstance(group, Group):
		group, created = Group.objects.get_or_create(name=group)
		if created:
			LOGGER.debug('Created new local group: %s', group)
			group_created.send_robust(sender=group)

	group.user_set.set(users)


def set_user_groups(user, groups):
	"""Replace the user groups
	Since the group system is not pluggable, any customization to it should be done elsewhere.

	:param user: the user instance
	:type user: object
	:param groups: a list of group instances or something that can be used to retrieve it
	:type groups: list[object|str]
	"""

	new_groups = {}
	for group in groups:
		if not isinstance(group, Group):
			group, created = Group.objects.get_or_create(name=group)
			if created:
				LOGGER.debug('Created new local group: %s', group)
				group_created.send_robust(sender=group)
		new_groups[group.name] = group

	LOGGER.debug('Updating groups for user: %s <- %s', user, new_groups.keys())
	user.groups.set(new_groups.values())



"""
Signal handlers. Should be included in the app's "ready" method and that would be it.
"""

from logging import getLogger

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.db.models.signals import m2m_changed
from django.dispatch import receiver

from .groups import user_joined_group, user_left_group
from ..utils import report_signal_results

LOGGER = getLogger(__name__)
UserModel = get_user_model()


@receiver(user_joined_group)
async def user_attribute_set_from_group(sender, **kwargs):
	"""User attribute from group join
	Use group join event to update the user's "is_superuser" and "is_staff" attributes.
	"""

	if not hasattr(settings, 'OKTA_CLIENT') or ('API' not in settings.OKTA_CLIENT):
		return

	if ((not kwargs['user'].is_superuser) or (not kwargs['user'].is_staff)) and ('SUPER_USER_GROUPS' in settings.OKTA_CLIENT['API']):
		if sender.name in settings.OKTA_CLIENT['API']['SUPER_USER_GROUPS']:
			LOGGER.debug('User "%s" is becoming a super user because of group membership: %s', kwargs['user'], sender.name)
			kwargs['user'].is_staff = True
			kwargs['user'].is_superuser = True
			await kwargs['user'].asave(update_fields=['is_staff', 'is_superuser'])
			return

	if not kwargs['user'].is_staff and ('STAFF_USER_GROUPS' in settings.OKTA_CLIENT['API']):
		if sender.name in settings.OKTA_CLIENT['API']['STAFF_USER_GROUPS']:
			LOGGER.debug('User "%s" is becoming a staff member because of group membership: %s', kwargs['user'], sender.name)
			kwargs['user'].is_staff = True
			await kwargs['user'].asave(update_fields=['is_staff'])


@receiver(user_left_group)
async def user_attribute_remove_from_group(sender, **kwargs):
	"""User attribute from group leave
	Use group leave event to update the user's "is_superuser" and "is_staff" attributes.
	"""

	if not hasattr(settings, 'OKTA_CLIENT') or ('API' not in settings.OKTA_CLIENT):
		return

	keep_staff = False
	if kwargs['user'].is_staff and ('STAFF_USER_GROUPS' in settings.OKTA_CLIENT['API']):
		if sender.name in settings.OKTA_CLIENT['API']['STAFF_USER_GROUPS']:
			current_groups = [user_group.name async for user_group in kwargs['user'].groups.all()]
			if frozenset(settings.OKTA_CLIENT['API']['SUPER_USER_GROUPS']) & frozenset(current_groups):
				keep_staff = True
			elif not kwargs['user'].is_superuser:
				LOGGER.debug('User "%s" is losing staff member status by leaving group: %s', kwargs['user'], sender.name)
				kwargs['user'].is_staff = False
				await kwargs['user'].asave(update_fields=['is_staff'])

	if (kwargs['user'].is_superuser or kwargs['user'].is_staff) and ('SUPER_USER_GROUPS' in settings.OKTA_CLIENT['API']):
		if sender.name in settings.OKTA_CLIENT['API']['SUPER_USER_GROUPS']:
			current_groups = [user_group.name async for user_group in kwargs['user'].groups.all()]
			if not (frozenset(settings.OKTA_CLIENT['API']['SUPER_USER_GROUPS']) & frozenset(current_groups)):
				LOGGER.debug('User "%s" is losing super user status by leaving group: %s', kwargs['user'], sender.name)
				kwargs['user'].is_staff = keep_staff
				kwargs['user'].is_superuser = False
				await kwargs['user'].asave(update_fields=['is_staff', 'is_superuser'])


@receiver(m2m_changed, sender=UserModel.groups.through)
async def signals_for_group_membership(sender, **kwargs):
	"""
	Trigger signals for group membership changes.
	"""

	results = []
	if kwargs['action'] == 'post_add':
		if kwargs['model'] is Group:
			groups = [await kwargs['model'].objects.aget(pk=pk) for pk in kwargs['pk_set']]
			for group in groups:
				results += await user_joined_group.asend_robust(sender=group, user=kwargs['instance'])
		elif kwargs['model'] is UserModel:
			users = [await kwargs['model'].objects.aget(pk=pk) for pk in kwargs['pk_set']]
			for user in users:
				results += await user_joined_group.asend_robust(sender=kwargs['instance'], user=user)
		else:
			LOGGER.error("Don't know how to handle user model to groups addition: %s", kwargs)

		# await sync_to_async(report_signal_results)(results, 'Group addition')

	elif kwargs['action'] == 'post_remove':
		if kwargs['model'] is Group:
			groups = [await kwargs['model'].objects.aget(pk=pk) for pk in kwargs['pk_set']]
			for group in groups:
				results += await user_left_group.asend_robust(sender=group, user=kwargs['instance'])
		elif kwargs['model'] is UserModel:
			users = [await kwargs['model'].objects.aget(pk=pk) for pk in kwargs['pk_set']]
			for user in users:
				results += await user_left_group.asend_robust(sender=kwargs['instance'], user=user)
		else:
			LOGGER.error("Don't know how to handle user model to groups removal: %s", kwargs)

		# await sync_to_async(report_signal_results)(results, 'Group removal')





#python
"""Okta Client Models
This module defines the Django models for integrating with Okta user profiles.
"""

from datetime import datetime as DateTime
from logging import getLogger

from django.contrib.auth.models import AbstractBaseUser, PermissionsMixin
from django.core.validators import RegexValidator
from django.db.models import BooleanField, CharField, DateTimeField, EmailField, TextChoices, URLField, fields
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from asgiref.sync import sync_to_async
from okta.exceptions.exceptions import OktaAPIException

from .api_client import OktaAPIClient, ERROR_CODE_MAP
from .groups import set_user_groups
from .managers import OktaUserManager

LOGGER = getLogger(__name__)

validator_5_to_100 = RegexValidator(regex = '^.{5,100}$', message = '5 <= value length <= 100')
nullable_5_to_100 = RegexValidator(regex = '(^.{5,100}$)|(ˆ$)', message = '5 <= value length <= 100 (or nothing)')


class OktaStatuses(TextChoices):
	"""Choices for OktaUser.okta_status
	Ref: https://help.okta.com/en-us/content/topics/users-groups-profiles/usgp-end-user-states.htm
	"""

	STAGED = 'STAGED', _('Staged')
	PROVISIONED = 'PROVISIONED', _('Pending User Action')
	ACTIVE = 'ACTIVE', _('Active')
	RECOVERY = 'RECOVERY', _('Password Reset')
	PASSWORD_EXPIRED = 'PASSWORD_EXPIRED', _('Password Expired')
	LOCKED_OUT = 'LOCKED_OUT', _('Locked Out')
	SUSPENDED = 'SUSPENDED', _('Suspended')
	DEPROVISIONED = 'DEPROVISIONED', _('Deprovisioned')


class AbstractOktaUser(AbstractBaseUser, PermissionsMixin):
	"""Default Okta profile
	Based on the official documentation as of (10/2021) https://developer.okta.com/docs/reference/api/users/#default-profile-properties
	"""

	login = CharField(primary_key=True, max_length=100, validators=[validator_5_to_100], verbose_name=_('login'), help_text=_('Unique identifier for the user (username)'))
	email = EmailField(blank=False, validators=[validator_5_to_100], verbose_name=_('email'), help_text=_('Primary email address of user'))
	secondEmail = EmailField(blank=True, validators=[nullable_5_to_100], verbose_name=_('second email'), help_text=_('Secondary email address of user typically used for account recovery'))
	firstName = CharField(blank=False, max_length=50, verbose_name=_('first name'), help_text=_('Given name of the user (givenName)'))
	lastName = CharField(blank=False, max_length=50, verbose_name=_('last name'), help_text=_('Family name of the user (familyName)'))
	middleName = CharField(blank=True, max_length=50, verbose_name=_('middle name'), help_text=_('Middle name(s) of the user'))
	honorificPrefix = CharField(blank=True, max_length=50, verbose_name=_('honorific prefix'), help_text=_('Honorific prefix(es) of the user, or title in most Western languages'))
	honorificSuffix = CharField(blank=True, max_length=50, verbose_name=_('honorific suffix'), help_text=_('Honorific suffix(es) of the user'))
	title = CharField(blank=True, max_length=100, verbose_name=_('title'), help_text=_('''User's title, such as "Vice President"'''))
	displayName = CharField(blank=True, max_length=250, verbose_name=_('display name'), help_text=_('Name of the user, suitable for display to end users'))
	nickName = CharField(blank=True, max_length=50, verbose_name=_('nickname'), help_text=_('Casual way to address the user in real life'))
	profileUrl = URLField(blank=True, verbose_name=_('profile URL'), help_text=_("URL of user's online profile (e.g. a web page)"))
	primaryPhone = CharField(blank=True, max_length=100, verbose_name=_('primary phone'), help_text=_('Primary phone number of user such as home number'))
	mobilePhone = CharField(blank=True, max_length=100, verbose_name=_('mobile phone'), help_text=_('Mobile phone number of user'))
	streetAddress = CharField(blank=True, max_length=100, verbose_name=_('street address'), help_text=_("Full street address component of user's address"))
	city = CharField(blank=True, max_length=100, verbose_name=_('city'), help_text=_("City or locality component of user's address (locality)"))
	state = CharField(blank=True, max_length=100, verbose_name=_('state'), help_text=_("State or region component of user's address (region)"))
	zipCode = CharField(blank=True, max_length=100, verbose_name=_('ZIP code'), help_text=_("Zipcode or postal code component of user's address (postalCode)"))
	countryCode = CharField(blank=True, max_length=2, verbose_name=_('country code'), help_text=_("Country name component of user's address (country)"))
	postalAddress = CharField(blank=True, max_length=100, verbose_name=_('postal address'), help_text=_("Mailing address component of user's address"))
	preferredLanguage = CharField(blank=True, max_length=100, verbose_name=_('preferred language'), help_text=_("User's preferred written or spoken languages"))
	locale = CharField(blank=True, max_length=5, verbose_name=_('locale'), help_text=_("User's default location for purposes of localizing items such as currency, date time format, numerical representations, etc."))
	timezone = CharField(blank=True, max_length=100, verbose_name=_('time zone'), help_text=_("User's time zone"))
	userType = CharField(blank=True, max_length=100, verbose_name=_('user type'), help_text=_('Used to describe the organization to user relationship such as "Employee" or "Contractor"'))
	employeeNumber = CharField(blank=True, max_length=100, verbose_name=_('employee number'), help_text=_('Organization or company assigned unique identifier for the user'))
	costCenter = CharField(blank=True, max_length=100, verbose_name=_('cost center'), help_text=_('Name of a cost center assigned to user'))
	organization = CharField(blank=True, max_length=100, verbose_name=_('organization'), help_text=_("Name of user's organization"))
	division = CharField(blank=True, max_length=100, verbose_name=_('division'), help_text=_("Name of user's division"))
	department = CharField(blank=True, max_length=100, verbose_name=_('department'), help_text=_("Name of user's department"))
	managerId = CharField(blank=True, max_length=100, verbose_name=_('manager ID'), help_text=_("ID of a user's manager"))
	manager = CharField(blank=True, max_length=100, verbose_name=_('manager'), help_text=_("DisplayName of the user's manager"))
	
	is_active = BooleanField(default=True, verbose_name=_("active"), help_text=_('Designates whether this user should be treated as active. \nUnselect this instead of deleting accounts.'))
	is_staff = BooleanField(default=False, verbose_name=_("staff status"), help_text=_("Designates whether the user can log into the admin site."))
	date_joined = DateTimeField(auto_now_add=True, verbose_name=_("date joined"), help_text=_('The timestamp when the local account was created'))

	okta_id = CharField(null=True, blank=True, unique=True, max_length=50, verbose_name=_('id'), help_text=_("Okta internal ID for the user."))
	okta_activated = DateTimeField(null=True, blank=True, verbose_name=_("activated"), help_text=_("Timestamp of the account activation."))
	okta_created = DateTimeField(null=True, blank=True, verbose_name=_("created"), help_text=_("Timestamp of the account creation."))
	okta_status = CharField(blank=True, max_length=30, choices=OktaStatuses, verbose_name=_('status'), help_text=_("Status of the Okta account."))
	okta_status_changed = DateTimeField(null=True, blank=True, verbose_name=_("status changed"), help_text=_('Timestamp of the last update of the "status" attribute.'))
	last_refresh_timestamp = DateTimeField(null=True, blank=True, verbose_name=_("Last refresh timestamp"), help_text=_('The last time the user was updated (refreshed) from Okta'))

	objects = OktaUserManager()
	
	EMAIL_FIELD = 'email'
	USERNAME_FIELD = 'login'
	REQUIRED_FIELDS = ['email', 'firstName', 'lastName']

	class Meta:
		verbose_name = _('okta user')
		verbose_name_plural = _('okta users')
		abstract = True

	def __getattr__(self, name):
		"""Lazy instantiation
		It provides a mechanism for lazy instantiation of the Okta API client.

		:param name: The name of the attribute being accessed.
		:type name: str
		:returns: the attribute value
		"""

		if name == '_api_client':
			value = OktaAPIClient()
		else:
			return getattr(super(), name)
		self.__setattr__(name, value)
		return value

	def __str__(self):
		"""String representation of the Okta user.

		:return: The login (username) of the Okta user.
		:rtype: str
		"""

		return self.login
	
	@classmethod
	def _attributes_from_okta_user(cls, okta_user):
		"""Extracts relevant attributes from an Okta user object.
		This method iterates through the model's fields and attempts to retrieve corresponding values from the provided `okta_user` object. It handles `DateTimeField` conversion and filters out empty or None values.

		:return: Attributes extracted from the Okta user, suitable for model instantiation or update.
		:rtype: dict
		"""

		attributes = {}
		for field in cls._meta.fields:
			if hasattr(okta_user.profile, field.name):
				okta_attr = getattr(okta_user.profile, field.name)
				if (okta_attr is None) or (isinstance(okta_attr, str) and not len(okta_attr)):
					continue
				if isinstance(field, fields.DateTimeField):
					okta_attr = DateTime.fromisoformat(okta_attr)
				attributes[field.name] = okta_attr

		attributes['okta_id'] = okta_user.id
		for dates_ in ('activated', 'created', 'status_changed'):
			if (attr := getattr(okta_user, dates_)) is not None:
				attributes[f'okta_{dates_}'] = DateTime.fromisoformat(attr)
		attributes['okta_status'] = OktaStatuses[okta_user.status]
		if attributes['okta_status'] != OktaStatuses.ACTIVE:
			attributes['is_active'] = False

		return attributes

	@classmethod
	def from_okta_user(cls, okta_user):
		"""Creates an instance of the OktaUser model from an Okta user object.
		This method uses `_attributes_from_okta_user` to extract relevant data and then instantiates the model.

		:param okta_user: An object representing the Okta user.
		:type okta_user: object
		:return: An instance of the OktaUser model.
		:rtype: cls
		"""

		return cls(**cls._attributes_from_okta_user(okta_user))

	def get_full_name(self):
		"""Returns the user's full name.
		This method prioritizes the `displayName` field. If `displayName` is not set, it constructs the full name by concatenating `honorificPrefix`, `firstName`, `middleName`, `lastName`, and `honorificSuffix`, including only non-empty components.

		:return: The user's full name
		:rtype: str
		"""

		if self.displayName:
			return self.displayName
		else:
			components = ('honorificPrefix', 'firstName', 'middleName', 'lastName', 'honorificSuffix')
			return ' '.join([getattr(self, component) for component in components if getattr(self, component)])

	def get_short_name(self):
		"""Returns the user's short name.
		This method prioritizes the `nickName` field. If `nickName` is not set, it returns the `firstName`.

		:return: The user's short name
		:rtype: str
		"""

		if self.nickName:
			return self.nickName
		else:
			return self.firstName

	@property
	def is_outdated(self):
		"""Are user attributes too old?
		Uses the "last_refresh_timestamp" and the API Client's "get_refresh_delta" to figure out if the attributes need to be updated.
		"""

		return (self.last_refresh_timestamp is None) or (now() > (self.last_refresh_timestamp + self._api_client.get_refresh_delta()))

	async def set_groups_from_okta(self, force_empty=False):
		"""Set groups from Okta
		Queries Okta for the group membership of the user and sets the local groups accordingly. Assumes that an empty list from "list_user_groups" was a failed query.

		:param force_empty: forces the update with an empty list if present (remove all groups)
		:type force_empty: bool
		"""

		if not self.okta_id:
			return
		okta_groups = await self._api_client('list_user_groups', self.okta_id)
		if okta_groups or force_empty:
			await sync_to_async(set_user_groups)(self, [group.profile.name async for group in okta_groups])

	def update(self, **updated_values):
		"""Updates user attributes from a dictionary of values.
		This method iterates through the provided keyword arguments and updates the corresponding fields on the user model instance. It only updates fields that exist on the model; any unknown fields will be ignored and a warning will be logged.

		:param updated_values: Keyword arguments where keys are model field names and values are the new values for those fields.
		"""
		
		local_fields = [field.name for field in self._meta.fields]
		for key, value in updated_values.items():
			if key in local_fields:
				setattr(self, key, value)
			else:
				LOGGER.warning('Dropping unknown field "%s" in object: %s', key, type(self))

	async def update_from_okta(self, force_update=False, save_model=True):
		"""Update attributes from Okta
		Queries the Okta API for the user's details and updates the local attributes
		"""

		if not await self._api_client.ping_users_endpoint():
			LOGGER.debug("Okta user endpoint not available; skipping user update: %s", self.login)
			return

		if self.is_outdated or force_update:
			if self.okta_id:
				okta_user = await self._api_client('get_user', self.okta_id)
			else:
				try:
					okta_user = await self._api_client('get_user', self.login)
				except OktaAPIException as error_:
					if error_.args[0]['errorCode'] != ERROR_CODE_MAP['USER_NOT_FOUND']:
						okta_user = None
					else:
						raise
				else:
					if self.login != okta_user.profile.login:
						okta_user = None
			if okta_user is None:
				LOGGER.debug('Local user not found in Okta: %s', self.login)
			else:
				LOGGER.debug('Retrieving info from Okta to update user: %s', self.login)
				await sync_to_async(self.update_from_okta_user)(okta_user, save_model=False)
				self.last_refresh_timestamp = now()
				if save_model:
					await self.asave()
		else:
			LOGGER.debug('User local attributes are current enough, skipping update: %s', self.login)

	def update_from_okta_user(self, okta_user, save_model=False):
		"""Updates the user's attributes from an Okta user object.
		This method extracts relevant attributes from the provided `okta_user` using `_attributes_from_okta_user` and then updates the corresponding fields on the current user instance.

		:param okta_user: An object representing the Okta user.
		:type okta_user: OktaAPIUser
		"""

		self.update(**self._attributes_from_okta_user(okta_user))
		if save_model:
			self.save()


class OktaUser(AbstractOktaUser):
	"""Django user model
	Alternate user model for Django based on Okta profiles.
	"""
	
	class Meta(AbstractOktaUser.Meta):
		swappable = 'AUTH_USER_MODEL'



#python
"""Okta user manager module
This module defines the custom user manager for the Okta user model.
"""

from logging import getLogger, INFO

from django.contrib.auth.base_user import BaseUserManager
from django.utils.timezone import now

from asgiref.sync import async_to_sync, sync_to_async
from okta.exceptions.exceptions import OktaAPIException
from okta.models.user import User as OktaAPIUser
from tqdm import tqdm as TQDM

from .api_client import OktaAPIClient
from .groups import set_group_members, set_user_groups

LOGGER = getLogger(__name__)


class OktaUserManager(BaseUserManager):
	"""Okta user remote manager
	Custom user manager to work with the Okta user model acting on the Okta directory via the API client.
	"""

	IDENTIFYING_FIELDS = ('pk', 'login', 'okta_id', 'email')

	use_in_migrations = True

	def __getattr__(self, name):
		"""Lazy instantiation
		It provides a mechanism for lazy instantiation of the Okta API client.

		:param name: The name of the attribute being accessed.
		:type name: str
		:returns: the attribute value
		"""

		if name == '_api_client':
			value = OktaAPIClient()
		else:
			return getattr(super(), name)
		self.__setattr__(name, value)
		return value

	def create_from_okta_user(self, okta_user):
		"""Create form Okta API User
		Creates a model object from an Okta API User.
		
		:param okta_user: the Okta User to base the model on
		:type okta_user: OktaAPIUser
		:return: the model object
		:rtype: self.model
		"""
		
		user = self.model.from_okta_user(okta_user)
		user.last_refresh_timestamp = now()
		user.set_unusable_password()
		user.save()
		return user
	
	def create_user(self, login, email, firstName, lastName, password=None, groups=None, **other_fields):
		"""Create a local user
		Create and save a user with the provided details.

		:param login: the pk of the model (username)
		:type login: str
		:param email: the email of the user
		:type email: str
		:param firstName: the first name of the user
		:type firstName: str
		:param lastName: the last name of the user
		:type lastName: str
		:param password: the password for the user, will be ignored
		:type password: str
		:param groups: a list of groups that the user will be added to
		:type groups: list
		:param other_fields: other fields passed to the user model
		:type other_fields: any
		"""

		try:
			email = self.normalize_email(email)
		except Exception:
			pass

		user = self.model(login=login, email=email, firstName=firstName, lastName=lastName, **other_fields)
		try:
			async_to_sync(user.update_from_okta)(save_model=False)
		except OktaAPIException as err:
			LOGGER.debug(f"Couldn't update the user {user.login} from Okta: {err}")

		user.set_unusable_password()
		user.save()

		try:
			async_to_sync(user.set_groups_from_okta)()
		except OktaAPIException as err:
			LOGGER.debug(f"Couldn't update user {user.login} groups from Okta: {err}")
		if groups is not None:
			set_user_groups(user=user, groups=groups)

		return user

	def create_superuser(self, login, email, firstName, lastName, password=None, groups=None, **other_fields):
		"""Create a local superuser
		Create and save a superuser with the provided details, including a password.

		:param login: the pk of the model (username)
		:type login: str
		:param email: the email of the user
		:type email: str
		:param firstName: the first name of the user
		:type firstName: str
		:param lastName: the last name of the user
		:type lastName: str
		:param password: the password for the user
		:type password: str
		:param groups: a list of groups that the user will be added to
		:type groups: list
		:param other_fields: other fields passed to the user model
		:type other_fields: any
		"""

		try:
			email = self.normalize_email(email)
		except Exception:
			pass

		other_fields.setdefault('is_staff', True)
		other_fields.setdefault('is_superuser', True)
		other_fields.setdefault('is_active', True)

		if other_fields.get('is_staff') is not True:
			raise ValueError('Superuser must have is_staff=True.')
		if other_fields.get('is_superuser') is not True:
			raise ValueError('Superuser must have is_superuser=True.')

		user = self.model(login=login, email=email, firstName=firstName, lastName=lastName, **other_fields)
		try:
			async_to_sync(user.update_from_okta)(save_model=False)
		except OktaAPIException as err:
			LOGGER.debug(f"Couldn't update the user {user.login} from Okta: {err}")
		if password is not None:
			user.set_password(password)
		user.save()

		try:
			async_to_sync(user.set_groups_from_okta)()
		except OktaAPIException as err:
			LOGGER.debug(f"Couldn't update user {user.login} groups from Okta: {err}")
		if groups is not None:
			set_user_groups(user=user, groups=groups)

		return user

	async def get_okta_updated(self, login, update_groups=True):
		"""Get user updated from Okta
		Returns a local model for a user with the attributes updated from Okta. If the user is not updated from Okta and is not an existing local user, it will return (None, False).

		:param login: the user login attribute (pk)
		:type login: str
		:return: the user
		:rtype: self.model
		"""

		user, created = await sync_to_async(self.get_or_create)(login=login)
		await user.update_from_okta()
		if created and not user.okta_id:
			await user.adelete()
			return None, False
		if update_groups:
			await user.set_groups_from_okta()

		return user

	async def update_all_from_okta(self, include_deprovisioned=True, include_groups=True, show_progress=False):
		"""Update all users and groups from Okta
		Queries Okta for the list of users, the list of groups, and the membership for each group. Updates the local users and groups accordingly.

		:param include_deprovisioned: also include users with the "Deprovisioned" status.
		:type include_deprovisioned: bool
		:param include_groups: also update groups.
		:type include_groups: bool
		:param show_progress: generate progress bar for the console output
		:type show_progress: bool
		:return: the number of users and groups updated
		:rtype: tuple[int, int]
		"""

		if show_progress:
			getLogger('asyncio').setLevel(INFO)

		okta_users = await self._api_client.list_users(include_deprovisioned=include_deprovisioned)
		last_total = len(okta_users)
		pbar_desc = 'Updating all Okta users' if include_deprovisioned else 'Updating non-deprovisioned Okta users'
		pbar = TQDM(desc=pbar_desc, total=last_total, disable=not show_progress, unit='users', dynamic_ncols=True)
		async for okta_user in okta_users:
			user, created = await self.aget_or_create(login=okta_user.profile.login)
			try:
				await sync_to_async(user.update_from_okta_user)(okta_user)
				await user.asave()
			except Exception:
				LOGGER.exception('User Okta update failed: %s', user)
			else:
				pbar.update()
			if last_total != len(okta_users):
				pbar.total = last_total = len(okta_users)
				pbar.refresh()
		pbar.close()

		okta_groups = {}
		if include_groups:
			okta_groups = await self._api_client('list_groups')

			last_total = len(okta_groups)
			pbar = TQDM(desc='Updating groups from Okta', total=last_total, disable=not show_progress, unit='groups', dynamic_ncols=True)
			async for okta_group in okta_groups:
				okta_members = await self._api_client('list_group_users', okta_group.id)
				try:
					group_members = [await self.aget(login=okta_member.profile.login) async for okta_member in okta_members]
				except self.model.DoesNotExist:
					LOGGER.error("Missing members prevented the update of group: %s <- %s", okta_group.profile.name, okta_members)
				else:
					await sync_to_async(set_group_members)(okta_group.profile.name, group_members)
				pbar.update()
				if last_total != len(okta_groups):
					pbar.total = last_total = len(okta_groups)
					pbar.refresh()

			pbar.close()

		return len(okta_users), len(okta_groups)



#!python3
"""
Custom command to update the local users by pulling the details via Okta API.
"""

from logging import getLogger

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand, CommandError

from asgiref.sync import async_to_sync
from okta.exceptions.exceptions import OktaAPIException

from ...api_client import OktaAPIClient

LOGGER = getLogger(__name__)

UserModel = get_user_model()


class Command(BaseCommand):
	"""
	The custom command
	"""

	help = "Update local users by pulling the details via Okta API"

	def add_arguments(self, parser):
		"""
		Adding some optional parameters
		"""

		parser.add_argument('--users', action='extend', nargs='+', help="A list of users that you'd like to update")
		parser.add_argument('--no-deprovisioned', action='store_true', help="Don't update deprovisioned users (when updating all users)")
		parser.add_argument('--no-groups', action='store_true', help="Don't update groups")

	def handle(self, *args, **options):
		"""Actual command behavior
		If a list of "users" is provided, only those will be updated; otherwise all of them will.
		"""

		api_client = OktaAPIClient()

		try:
			async_to_sync(api_client.ping_users_endpoint)(required=True)
		except (AttributeError, KeyError):
			raise CommandError("The API client doesn't seem to be configured")
		except OktaAPIException as error_:
			raise CommandError(f'There seems to be an issue with the Okta configuration; attempting to access the list of users yield: {error_}')
		self.stdout.write('The Okta client is configured and functional')

		success_count, group_count, failures = 0, 0, []
		if ('users' in options) and options['users']:
			self.stdout.write(f"Updating {len(options['users'])} users")
			groups = set()
			for user in options['users']:
				try:
					user = UserModel.objects.get(login=user)
				except UserModel.DoesNotExist:
					failures.append(user)
					continue
				async_to_sync(user.update_from_okta)(force_update=True)
				success_count += 1
				if not options['no_groups']:
					async_to_sync(user.set_groups_from_okta)()
					groups |= frozenset([group.name for group in user.groups.all()])
			group_count = len(groups)
		else:
			success_count, group_count = async_to_sync(UserModel.objects.update_all_from_okta)(include_deprovisioned=not options['no_deprovisioned'], include_groups=not options['no_groups'], show_progress=True)

		if failures:
			self.stdout.write(self.style.ERROR(f"Failed to update {len(failures)} users: {', '.join(failures)}"))
			if success_count:
				self.stdout.write(f'The other {success_count} users were updated')
			else:
				self.stdout.write(self.style.WARNING('No users were updated'))
		else:
			self.stdout.write(self.style.SUCCESS(f'Successfully updated {success_count} users'))

		if group_count:
			self.stdout.write(self.style.NOTICE(f'Updated {group_count} groups from Okta with unknown results (check the output for errors)'))
		else:
			self.stdout.write(self.style.NOTICE('No groups were updated'))



"""
Here are some utility pieces.
"""

from collections.abc import Sequence
from datetime import timedelta as TimeDelta
from logging import getLogger

from django.conf import settings
from django.core.cache import cache

from okta.client import Client as OktaClient

ERROR_CODE_MAP = {
	'USER_NOT_FOUND' : 'E0000007',
}
LOGGER = getLogger(__name__)


class OktaResultCollection:
	"""Okta result collection
	Custom generator for Okta API queries.
	"""

	def __init__(self, results, response, buffer_size=25):
		"""Magic initialization
		It expects the results of the initial query.

		:param results: The results of the initial query.
		:type results: list[any]
		:param response: The response from the initial query.
		:type response: oktaResponse
		:param buffer_size: The size of the buffer.
		:type buffer_size: int
		"""

		if not isinstance(results, Sequence):
			raise ValueError('result must be a sequence')

		self._current_response = response
		self._current_results = results
		self._current_results.reverse()
		self._buffer_size = buffer_size

		self._entries = []
		self._total = len(self._current_results)

	def __aiter__(self):
		"""Asynchronous iterator
		This class itself is an asynchronous iterator.

		:return: self
		:rtype: OktaResultCollection
		"""

		return self

	async def __anext__(self):
		"""Asynchronous next
		Pops one of the values from self._current_results. It includes logic to load more values (paginated API response) and to signal the exhaustion.

		:return: the "next" result
		:rtype: any
		"""

		if len(self._current_results) < self._buffer_size:
			try:
				await self._load_results()
			except StopAsyncIteration:
				pass

		if not self._current_results:
			raise StopAsyncIteration

		return self._current_results.pop()

	def __len__(self):
		"""Length
		Implement the length protocol, unsupported in regular generators.
		"""

		return self._total

	async def _load_results(self):
		"""Load more results
		Lightly convoluted logic to retrieve extra results.
		1. get the results from the next page of the request, if there's one
		2. move to the next "request" and do the same

		:return: the number of results loaded
		:rtype: int
		"""

		if self._current_response.has_next():
			results, err = await self._current_response.next()
			results.reverse()
			self._current_results = results + self._current_results
			self._total += len(results)
			return len(results)
		else:
			if self._entries:
				results, response = self._entries.pop()
				self._current_response = response
				results.reverse()
				self._current_results = results + self._current_results
				self._total += len(results)
				return len(results)
			else:
				return 0

	def append(self, *others):
		"""Append request
		Include another request to the end of the collection.
		"""

		for other in others:
			if not isinstance(other, type(self)):
				raise ValueError(f'Can only append {type(self)}')
			results, response = other._current_results, other._current_response
			results.reverse()
			self._entries = other._entries + [[results, response]] + self._entries


class OktaAPIClient:
	"""
	Handles interactions with the Okta API, including lazy instantiation of the Okta client and making API requests.
	"""

	CACHE_PREFIX = 'okta_client_api_client'
	STATIC_CONFIG = {'raiseException': True}

	def __getattr__(self, name):
		"""Lazy instantiation
		It provides a mechanism for lazy instantiation of the Okta API client, its credentials, and the Okta organization URL.

		:param name: The name of the attribute being accessed.
		:type name: str
		:returns: the attribute value
		"""

		if name == 'okta_api_client':
			client_config = {'orgUrl': settings.OKTA_CLIENT['API']['ORG_URL']} | self.okta_api_credentials
			if hasattr(settings, 'SSL_CONTEXT'):
				client_config['sslContext'] = settings.SSL_CONTEXT
			value = OktaClient(client_config | self.STATIC_CONFIG)
		elif name == 'okta_api_credentials':
			if ('API_CLIENT_ID' in settings.OKTA_CLIENT['API']) and ('API_PRIVATE_KEY' in settings.OKTA_CLIENT['API']):
				value = {
					'authorizationMode'	: 'PrivateKey',
					'clientId'			: settings.OKTA_CLIENT['API']['API_CLIENT_ID'],
					'privateKey'		: settings.OKTA_CLIENT['API']['API_PRIVATE_KEY'],
					'scopes'			: settings.OKTA_CLIENT['API'].get('API_SCOPES', None),
				}
			elif 'API_TOKEN' in settings.OKTA_CLIENT['API']:
				value = {'token': settings.OKTA_CLIENT['API']['API_TOKEN']}
			else:
				raise RuntimeError('Missing auth settings for Okta client')
		else:
			return getattr(super(), name)
		self.__setattr__(name, value)
		return value

	async def __call__(self, method_name, *args, retrieve_all_pages=True, **kwargs):
		"""Okta API request
		Makes a request to the Okta API using the configured Okta client.

		:param method_name: The name of the method to call on the Okta client (e.g., 'list_users').
		:type method_name: str
		:param args: Positional arguments to pass to the Okta client method.
		:param kwargs: Keyword arguments to pass to the Okta client method.
		:return: The result of the Okta API call.
		"""

		result = await getattr(self.okta_api_client, method_name)(*args, **kwargs)

		if len(result) == 3:
			result, response, err = result
		else:
			raise RuntimeError('Unknown result: {}'.format(result))

		if isinstance(result, Sequence):
			return OktaResultCollection(results=result, response=response)
		else:
			return result
	
	@staticmethod
	def get_refresh_delta():
		"""Get refresh delta
		Load the USER_TTL setting and create the equivalent timedelta object.

		:return: a timedelta object representing the configured TTL time
		:rtype: timedelta
		"""
		
		if hasattr(settings, 'OKTA_CLIENT') and ('API' in settings.OKTA_CLIENT) and ('USER_TTL' in settings.OKTA_CLIENT['API']):
			return TimeDelta(seconds=settings.OKTA_CLIENT['API']['USER_TTL'])
		else:
			return TimeDelta(seconds=0)

	async def list_users(self, include_deprovisioned=False, **kwargs):
		"""List all users
		A subset of users can be returned that match a supported filter expression or search criteria. Different results are returned depending on specified queries in the request.
		It will swallow the exceptions and report via logging, for your convenience.

		:param include_deprovisioned: adds a second query to also fetch deprovisioned users
		:type include_deprovisioned: bool
		:param kwargs: any filters or otherwise that will be passed as is to "list_users"
		:type kwargs: any
		:return: a list of users
		:rtype: list[OktaAPIUser]
		"""

		await self.ping_users_endpoint(required=True)

		result = await self('list_users', query_params=kwargs)

		if include_deprovisioned:
			if 'search' in kwargs:
				raise ValueError("Not clear how to merge filter into query")
			else:
				kwargs['search'] = 'status eq "Deprovisioned"'

			result.append(await self('list_users', query_params=kwargs))

		return result

	async def ping_users_endpoint(self, required=False):
		"""Ping users endpoint
		Attempt a query to the users endpoint and report availability.

		:return: True if the query succeeds.
		:rtype: bool
		"""

		try:
			return len(await self('list_users', retrieve_all_pages=False, query_params={'limit':'1'})) > 0
		except Exception:
			if required:
				raise
			else:
				return False


#python
"""
Django Okta Client Views
"""

from json import JSONDecodeError
from logging import getLogger

from django.contrib.auth.mixins import LoginRequiredMixin
from django.http import HttpResponse, HttpResponseBadRequest, HttpResponseRedirect, JsonResponse
from django.shortcuts import render
from django.views import View

from asgiref.sync import async_to_sync
from rest_framework.permissions import IsAuthenticated as RESTIsAuthenticated
from rest_framework.views import APIView

from .exceptions import SAMLAssertionError
from .mixins import LoginLogoutMixin, OktaEventHookMixin

LOGGER = getLogger(__name__)


class ACSView(LoginLogoutMixin, View):
	"""
	Handles the Assertion Consumer Service (ACS) for SAML responses from Okta.
	"""

	def post(self, request):
		"""POST verb
		Handles the POST request for the ACS endpoint, processing the SAML assertion.

		:param request: the Django request
		:type request: object
		:returns object: the Django response
		"""

		try:
			next_url = self.saml_assertion(request)
		except SAMLAssertionError:
			return HttpResponseBadRequest()
		else:
			return HttpResponseRedirect(next_url)


class LoginView(LoginLogoutMixin, View):
	"""
	Handles the login process for Okta.
	"""

	def get(self, request):
		"""GET verb
		Handles the GET request for the login endpoint.

		:param request: the Django request
		:type request: object
		:returns object: the Django response
		"""

		try:
			next_url = self.login_user(request)
		except RuntimeError:
			return HttpResponseBadRequest()
		else:
			return HttpResponseRedirect(next_url)


class LogoutView(LoginLogoutMixin, LoginRequiredMixin, View):
	"""
	Handles the logout process for Okta.
	"""

	def get(self, request):
		"""GET verb
		Handles the GET request for the logout endpoint.

		:param request: the Django request
		:type request: object
		:returns object: the Django response
		"""

		next_url = self.logout_user(request)
		return HttpResponseRedirect(next_url)

	def post(self, request):
		"""POST verb
		Handles the POST request for the logout endpoint.

		:param request: the Django request
		:type request: object
		:returns object: the Django response
		"""

		next_url = self.logout_user(request)
		return HttpResponseRedirect(next_url)


class OktaEventHooks(OktaEventHookMixin, APIView):
	"""
	Handles Okta event hooks.
	"""

	permission_classes = [RESTIsAuthenticated]

	def get(self, request):
		"""HTTP GET
		Only used to confirm that it follows Okta's convention.

		:param request: the Django request
		:type request: object
		:returns object: the Django response
		"""

		return JsonResponse(self.authenticate_endpoint(request))

	def post(self, request):
		"""HTTP GET
		Regular Event Hook handling.

		:param request: the Django request
		:type request: object
		:returns object: the Django response
		"""

		try:
			async_to_sync(self.handle_events)(request)
		except (JSONDecodeError, UnicodeDecodeError):
			return HttpResponseBadRequest()
		else:
			return HttpResponse(status=204)


######## Demo View ########

class IndexView(LoginRequiredMixin, View):
	"""
	A simple view to demonstrate a protected page after successful login.
	"""

	def get(self, request):
		"""GET verb
		Handles the GET request for the index page.

		:param request: the Django request
		:type request: object
		:returns object: the Django response
		"""

		return render(request, 'okta-client/index.html')


"""Local signal handlers
These are available to be called before the respective signals are triggered.
"""

from logging import getLogger

from django.contrib.auth import get_user_model

LOGGER = getLogger(__name__)

UserModel = get_user_model()


async def update_user_from_okta(request, event):
	"""Update user from Okta
	An adaptor from the signal handling code to the UserModel manager's "get_user".
	"""

	for target in event['target']:
		await UserModel.objects.get_okta_updated(login=target['alternateId'], update_groups=False)



"""
This module defines Django signals related to Django groups.
"""

from logging import getLogger

from django.dispatch import Signal

LOGGER = getLogger(__name__)

group_created = Signal()
user_joined_group = Signal()
user_left_group = Signal()
